from __future__ import print_function
import sys, webbrowser, traceback
from tracestack.engines import GoogleEngine, StackEngine
from tracestack.utils import getch

class ExceptionHandler(object):
    """Callable exception handler that can replace sys.__excepthook__."""

    def __init__(self, skip=False, engine="default", *args, **kwargs):
        """Initializer takes same arguments as pm, enable, trace, etc.

        Args:
            skip (bool) -- whether to skip the prompt (default: False)
            engine (string) -- the search engine to use (default: "default")
                'default': Google limited to stackoverflow.com, 
                'google': full web search on Google, 
                'stackoverflow': StackOverflow site search
        """
        if engine in ("default", "google"):
            self.engine = GoogleEngine(engine=engine, *args, **kwargs)
        elif engine == "stackoverflow":
            self.engine = StackEngine(*args, **kwargs)
        else:
            msg = "'%s' is not a valid engine option (choose between " + \
                  "'default', 'google', and 'stackoverflow')"
            raise ValueError(msg % engine)
        self.skip = skip

    def __call__(self, *einfo):
        """Handles error.  Takes same three arguments as 
        sys.__excepthook__: type, value, traceback."""
        einfo = einfo or sys.exc_info()
        self._print_traceback(*einfo)
        self.handle_error(*einfo)

    def handle_error(self, *einfo):
        error_string = self._get_error_string(*einfo)
        self._handle_string(error_string)

    def _print_traceback(self, *einfo):
        traceback.print_exception(*einfo)

    def _get_error_string(self, *einfo):
        (etype, evalue, tb) = einfo
        error_string = traceback.format_exception_only(etype, evalue)[-1]
        return error_string

    def _search(self, error_string):
        search_url = self.engine.search(error_string)
        webbrowser.open(search_url)

    def _prompt(self):
        if self.skip:
            return True
        else:
            choice = getch("Hit spacebar to search this error message on %s: " % self.engine.name())
            if choice == " ":
                return True
            else:
                return False

    def _handle_string(self, error_string):
        if self._prompt():
            self._search(error_string)



import sys
from tracestack.handler import ExceptionHandler

def pm(*args, **kwargs):
    """Post-mortem function that searches the last exception."""

    if hasattr(sys, "last_type"):
        einfo = (sys.last_type, sys.last_value, sys.last_traceback)
        ExceptionHandler(*args, **kwargs)(*einfo)
    else:
        raise ValueError("no last exception")

def enable(*args, **kwargs):
    """Install the tracestack exception handler as the system exception 
    handler."""

    sys.excepthook = ExceptionHandler(*args, **kwargs)

def disable(*args, **kwargs):
    """Revert to the default exception handler."""

    sys.excepthook = sys.__excepthook__




"""Gets a single character from the user.  Works cross-platform.

Thanks to Danny Yoo (http://code.activestate.com/recipes/134892/)
"""
from __future__ import print_function
import sys

class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self, msg):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self, msg):
        import msvcrt
        return msvcrt.getch()


class _Getch:
    """Gets a single character from standard input.  Does not echo to the
    screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self, msg): 
        print(msg, end="")
        sys.stdout.flush()
        char = self.impl(msg)
        print(char)
        return char



getch = _Getch()


import unittest
import sys
import tracestack
import webbrowser

try:
	from mock import Mock, patch
except ImportError:
	from unittest.mock import Mock, patch

exc_return_value = (type(Exception("exception message")), Exception("exception message"), Mock())
mock_sys_exc_info = Mock(return_value=exc_return_value)
mock_webbrowser = Mock()
mock_traceback_print_exception = Mock()
mock_getch = Mock(return_value=" ")
mock_sys = Mock()
mock_sys.last_type = exc_return_value[0]
mock_sys.last_value = exc_return_value[1]
mock_sys.last_traceback = exc_return_value[2]


@patch('tracestack.handler.sys.exc_info', mock_sys_exc_info)
@patch('tracestack.handler.webbrowser', mock_webbrowser)
@patch('tracestack.handler.traceback.print_exception', mock_traceback_print_exception)
@patch('tracestack.handler.getch', mock_getch)
class TestHandler(unittest.TestCase):

	default_url = 		('http://www.google.com/search?q=Exception%3A' + 
						 '+exception+message%0A+python+site%3Astackoverflow.com' +
						 '+inurl%3Aquestions')
	google_url = 		('http://www.google.com/search?q=Exception%3A' + 
						 '+exception+message%0A+python')
	stackoverflow_url = ('http://www.stackoverflow.com/search?' + 
						 'q=Exception%3A+exception+message%0A+%5Bpython%5D')
	default_prompt = 	   'Hit spacebar to search this error message on Stack Overflow (using Google): '
	google_prompt =		   'Hit spacebar to search this error message on the web (using Google): '
	stackoverflow_prompt = 'Hit spacebar to search this error message on Stack Overflow: '

	def setUp(self):
		pass

	def tearDown(self):
		mock_sys_exc_info.reset_mock()
		mock_webbrowser.reset_mock()
		mock_traceback_print_exception.reset_mock()
		mock_getch.reset_mock()

	def test_default(self):
		handler = tracestack.handler.ExceptionHandler()
		handler()
		self.check_results()

	def test_skip(self):
		handler = tracestack.handler.ExceptionHandler(skip=True)
		handler()
		self.skip=True
		self.check_results(skip=True)

	def test_google(self):
		handler = tracestack.handler.ExceptionHandler(engine="google")
		handler()
		self.check_results(prompt=self.google_prompt, url=self.google_url)

	def test_stackoverflow(self):
		handler = tracestack.handler.ExceptionHandler(engine="stackoverflow")
		handler()
		self.check_results(prompt=self.stackoverflow_prompt, url=self.stackoverflow_url)

	def check_results(self, skip=False, prompt=None, 
					  url=None):
		prompt = prompt or self.default_prompt
		url = url or self.default_url
		mock_traceback_print_exception.assert_called_once_with(*exc_return_value)
		if skip:
			mock_getch.assert_not_called()
		else:
			mock_getch.assert_called_with(prompt)
		mock_webbrowser.open.assert_called_once_with(url)

	@patch('tracestack.debugger.sys', mock_sys)
	def test_pm(self):
		tracestack.pm()
		self.check_results()

	def test_nothing(self):
		with self.assertRaises(AssertionError):
			self.check_results()		


class TestFunctions(unittest.TestCase):

	def test_enable(self):
		tracestack.enable()
		self.assertIsNot(sys.excepthook, sys.__excepthook__)
		self.assertTrue(isinstance(sys.excepthook, 
								   tracestack.handler.ExceptionHandler))
		tracestack.disable()
		self.assertIs(sys.excepthook, sys.__excepthook__)

	def test_decorator(self):
		@tracestack.trace
		def buggy_function():
			self.assertIsNot(sys.excepthook, sys.__excepthook__)
			self.assertTrue(isinstance(sys.excepthook, 
									   tracestack.handler.ExceptionHandler))
		buggy_function()
		self.assertIs(sys.excepthook, sys.__excepthook__)





import sys
from functools import wraps
from tracestack.handler import ExceptionHandler

def trace(*args, **kwargs):
    """Decorator that applies the tracestack exception handler to one function.
    Optionally takes the same arguments as the other functions.

    Examples:
    
    @tracestack.trace
    def buggy_function():
        ...

    @tracestack.trace(skip=True)
    def buggy_function():
        ...

    """

    def decorator(func):
        """The decorator itself."""
        @wraps(func)
        def wrapper(*args, **kwargs):
            sys.excepthook = handler
            result = func(*args, **kwargs)
            sys.excepthook = sys.__excepthook__
            return result
        return wrapper

    if len(args) == 1 and callable(args[0]):
        # @tracestack was used as a decorator without arguments.
        # Return the decorated function.
        handler = ExceptionHandler()
        return decorator(args[0])
    else:
        # @tracestack(...) was called with arguments.
        # Return a decorator based on those arguments.
        handler = ExceptionHandler(*args, **kwargs)
        return decorator



import sys
from code import InteractiveConsole
from tracestack.handler import ExceptionHandler


class TracestackConsole(InteractiveConsole):
    """An interactive console for running code or REPL loops that handle
    errors using the tracestack handler."""

    def __init__(self, *args, **kwargs):
        self.handler = ExceptionHandler(*args, **kwargs)
        kwargs.pop("engine")
        kwargs.pop("skip")
        kwargs.pop("arguments")
        InteractiveConsole.__init__(self, *args, **kwargs)

    def showtraceback(self):
        InteractiveConsole.showtraceback(self)
        self.handler.handle_error(*sys.exc_info())

    def showsyntaxerror(self, filename=None):
        InteractiveConsole.showsyntaxerror(self, filename)
        self.handler.handle_error(*sys.exc_info())


from __future__ import print_function
import sys
import os
import runpy
import traceback
import contextlib
import argparse
from tracestack.handler import ExceptionHandler
from tracestack.console import TracestackConsole

def run():
    """Runs the script provided by the arguments, using the 
    tracestack exception handler.
    """
    parser = _build_parser()
    args = vars(parser.parse_args())
    print(args)
    script = args.pop("script")
    handler = ExceptionHandler(**args)
    if script:
        # set up the system variables
        sys.argv = sys.argv[1:]
        sys.path[0] = os.path.dirname(os.path.abspath(script))
        try:
            runpy.run_path(script, run_name="__main__")
        except:
            einfo = sys.exc_info()
            _print_clean_traceback(einfo)
            handler.handle_error(*einfo)
    else:
        # no additional arguments were given; run the REPL loop
        console = TracestackConsole(**args)
        console.interact()

def _build_parser():
    """Returns the argument parser (which is built using argparse)."""
    
    parser = argparse.ArgumentParser(description='instantly search your ' + \
                                                 'Python error messages on' + \
                                                 ' the web',
                                    formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('script', metavar='SCRIPT', type=str, nargs='?',
                        help='the Python script')
    parser.add_argument('arguments', metavar='ARGUMENTS', type=str, nargs='*',
                        help='any arguments to the script')
    parser.add_argument('-s', '--skip', help='skip the prompt and immediately search each exception',
                        action='store_true')
    parser.add_argument('-e', 
                        '--engine', 
                        help="""the search engine to use:
  'default': Google limited to stackoverflow.com, 
  'google': full web search on Google, 
  'stackoverflow': StackOverflow site search""", 
                        default="default", 
                        choices=['default', 'google', 'stackoverflow'], 
                        metavar="ENGINE",
                        type=str)
    return parser

def _print_clean_traceback(einfo):
    """Print the traceback, without showing all the overhead added by tracestack."""

    extracted = traceback.extract_tb(einfo[2])
    if extracted[-1][0] in ('trace', 'runpy.py'):
        # the error call is coming from inside the house
        # this shouldn't happen, but if it does, do the default behavior
        sys.__excepthook__(sys.exc_info)
    else:
        # remove the traceback levels that relate to runpy or trace
        extracted = [level for level in 
                     extracted if 
                     os.path.basename(level[0]) not in ('command_line.py', 'runpy.py')]
        extracted = traceback.format_list(extracted)
        # print as if it were a normal traceback
        print("Traceback (most recent call last):", file=sys.stderr)
        for level in extracted:
            print(level, end="", file=sys.stderr)
        for line in traceback.format_exception_only(einfo[0], einfo[1]):
            print(line, end="", file=sys.stderr)
